# Generated by Django 4.1.7 on 2023-02-22 21:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('nets_core', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='verificationcode',
            name='token',
            field=models.CharField(max_length=150, verbose_name='Verification token'),
        ),
    ]



from datetime import timedelta
from django.core.mail import EmailMultiAlternatives
from django.db.utils import IntegrityError
from django.http import HttpResponse
from django.shortcuts import render
import logging
from django.utils.translation import gettext_lazy as _
from nets_core.decorators import request_handler
from nets_core.models import UserDevice, VerificationCode
from nets_core.params import RequestParam
from nets_core.responses import error_response, success_response
from nets_core.security import authenticate
from django.contrib.auth import login, logout
from django.utils import timezone
from django.conf import settings
from oauth2_provider.models import Application, AccessToken, RefreshToken
from oauthlib import common
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured

User = get_user_model()
username_field = getattr(User, "USERNAME_FIELD", "username")

logger = logging.getLogger(__name__)

prohibited_fields = [
    "password",
    "is_superuser",
    "is_staff",
    "is_active",
    "verified",
    "email_verified",
    "last_login",
    "date_joined",
    "updated_fields",
    "groups",
    "user_permissions",
    "doc_*",
]
try:
    if settings.NETS_CORE_USER_PROHIBITED_FIELDS:
        prohibited_fields += settings.NETS_CORE_USER_PROHIBITED_FIELDS
except:
    pass


def valid_gender(s):
    return s in ["male", "female", "other", "_"]


@request_handler(
    public=True,
    params=[
        RequestParam(username_field, str),
        RequestParam("device", dict, True, default=None),
    ],
)
def auth_login(request):
    print(request.params)
    try:
        defaults = {}
        for key, val in request.params._asdict().items():
            if hasattr(User, key):
                for field in prohibited_fields:
                    if field in key:
                        # exact match e.g. password
                        logger.info(f"Prohibited field {key} found in auth request")
                        continue
                    if field.endswith("*"):
                        # check if key starts with field
                        # e.g. doc_id, doc_id_type, doc_id_country
                        if key.startswith(field[:-1]):
                            logger.info(f"Prohibited field {key} found in auth request")
                            continue

                defaults[key] = val

        new_user, created = User.objects.get_or_create(
            **{username_field: getattr(request.params, username_field)},
            defaults=defaults,
        )

        device = None

        if hasattr(request.params, "device") and request.params.device is not None:
            valid_device_fields = [
                "name",
                "os",
                "os_version",
                "device_token",
                "firebase_token",
                "app_version",
                "device_id",
                "device_type",
            ]
            device_data = {}

            for key, val in request.params.device.items():
                if key in valid_device_fields:
                    device_data[key] = val

            if "uuid" in request.params.device and request.params.device["uuid"]:
                logger.info(
                    f"Device uuid {request.params.device['uuid']} found in auth request"
                )
                try:
                    device = UserDevice.objects.get(
                        uuid=request.params.device["uuid"], user=new_user
                    )

                    for key, val in device_data.items():
                        setattr(device, key, val)
                    device.save()

                    # TODO: Notification to user to new login from device

                except UserDevice.DoesNotExist:
                    # uuid does not exist or is not associated with user
                    # delete user if created as invalid request is made
                    logger.warning(
                        f"Invalid device uuid {request.params.device['uuid']} found in auth request from user request {request.params}"
                    )
                    if created:
                        new_user.delete()

                    return error_response(_("Invalid device uuid"), 400)
            else:
                # new device
                device_data["user"] = new_user
                if "firebase_token" in device_data:
                    device = UserDevice.objects.filter(
                        firebase_token=device_data["firebase_token"], user=new_user
                    )
                    if device.exists():
                        device = device.first()
                        for key, val in device_data.items():
                            setattr(device, key, val)
                        device.save()
                    else:
                        device = UserDevice.objects.create(**device_data)
                else:
                    device = UserDevice.objects.create(**device_data)

        # create verification code, listeners will send email/sms and devices notifications
        # with firebase
        VerificationCode.objects.create(user=new_user, ip=request.ip, device=device)

        return success_response(
            _("CODE SENT"),
            extra={
                "device_uuid": device.uuid if device else None,
            },
        )

    except IntegrityError as e:
        print(e)
        return error_response(_("Username not available"), 400)

    except Exception as e:
        msg = e.__str__()
        return error_response(msg)


@request_handler(params=[RequestParam("email", "email")], public=True)
def check_email(request):
    print(request.params.email)
    exists = False
    print(exists)
    return success_response(exists)


@request_handler(
    User,
    index_field=username_field,
    params=[
        RequestParam(username_field, str),
        RequestParam("code", int),
        RequestParam("client_id", str),
        RequestParam("client_secret", str),
    ],
    public=True,
)
def auth(request):
    user = request.obj
    device_uuid = None
    if hasattr(request.params, "device_uuid"):
        try:
            device = UserDevice.objects.get(uuid=request.params.device_uuid, user=user)
            device_uuid = device.uuid
        except UserDevice.DoesNotExist:
            return error_response(_("Invalid device uuid"), 400)

    try:
        tokens = authenticate(
            user=user,
            code=request.params.code,
            client_id=request.params.client_id,
            client_secret=request.params.client_secret,
            device_uuid=device_uuid,
        )
        if hasattr(user, "email_verified") and not user.email_verified:
            user.email_verified = True
        user.last_login = timezone.now()
        user.save()
        tokens.update({"user": user.to_json()})
        # add user to session
        login(request, user, backend="django.contrib.auth.backends.ModelBackend")
        return success_response(tokens)

    except Exception as e:
        msg = e.__str__()
        return error_response(msg)


@request_handler()
def auth_logout(request):
    params = request.params
    if hasattr(params, "device_uuid"):
        try:
            device = UserDevice.objects.get(uuid=params.device_uuid, user=request.user)
            device.delete()
        except UserDevice.DoesNotExist:
            pass
    # get Bearer token
    try:
        access_token = request.headers["Authorization"].split(" ")[1]
        token = AccessToken.objects.get(token=access_token)
        token.delete()
        refresh_token = RefreshToken.objects.get(access_token=token)
        refresh_token.delete()
    except Exception as e:
        pass

    logout(request)
    return success_response(_("Logged out successfully"))


@request_handler(
    params=[
        RequestParam("fields", list, True, default=None),
    ]
)
def auth_get_profile(request):
    request.user.save()
    fields = request.params.fields 
    # convert list in tuple
    if fields:
        fields = tuple(fields)   
        
    return success_response(request.user.to_json(fields=fields))


@request_handler()
def update_user(request):
    user = request.user
    updated_fields = {}

    if hasattr(user, "updated_fields"):
        updated_fields = user.updated_fields

    for key, val in request.params._asdict().items():
        if hasattr(user, key):
            for field in prohibited_fields:
                if field in key:
                    # exact match e.g. password
                    logger.info(f"Prohibited field {key} found in auth request")
                    continue

                if field.endswith("*"):
                    # check if key starts with field
                    # e.g. doc_id, doc_id_type, doc_id_country
                    if key.startswith(field[:-1]):
                        # logger.info(f"Prohibited field {key} found in auth request")
                        continue

            if key not in updated_fields:
                updated_fields[key] = []
            updated_fields[key].append(
                {
                    "old": str(getattr(user, key)),
                    "new": str(val),
                    "time": timezone.now().__str__(),
                }
            )
            setattr(user, key, val)

    if request.FILES:
        for field in request.FILES:
            if not hasattr(user, field):
                continue

            for field in prohibited_fields:
                if field in key:
                    # exact match e.g. password
                    logger.info(f"Prohibited field {key} found in auth request")
                    continue

                if field.endswith("*"):
                    # check if key starts with field
                    # e.g. doc_id, doc_id_type, doc_id_country
                    if key.startswith(field[:-1]):
                        # logger.info(f"Prohibited field {key} found in auth request")
                        continue

                if field not in updated_fields:
                    updated_fields[field] = []
                updated_fields[field].append(
                    {
                        "old": str(getattr(user, field)),
                        "new": str(request.FILES[field]),
                        "time": timezone.now().__str__(),
                    }
                )

                setattr(user, field, request.FILES[field])

    user.updated_fields = updated_fields
    try:
        user.save()
    except Exception as e:
        return error_response(e.__str__())

    return success_response(user.to_json())


@request_handler(public=True)
def request_delete_user_account(request):
    if not hasattr(settings, "NETS_CORE_DELETE_ACCOUNT_TEMPLATE"):
        info_template = "nets_core/delete_account_info.html"
    else:
        info_template = settings.NETS_CORE_DELETE_ACCOUNT_TEMPLATE

    return render(
        request,
        "nets_core/delete_account.html",
        {
            "title": _("Delete Account"),
            "info_template": info_template,
            "user": request.user,
        },
    )


@request_handler(
    params=[RequestParam("sure", bool, default=False), RequestParam("code", str)],
)
def delete_user_account(request):
    if not request.params.sure:
        return error_response(_("Are you sure?"), 400)

    if not request.params.code:
        return error_response(_("Invalid code"), 400)

    try:
        code = (
            VerificationCode.objects.get(user=request.user, verified=False)
            .order_by("-created")
            .first()
        )

        if code.validate(request.params.code):
            # delete user
            user = request.user
            user.delete()
            return success_response(_("Account deleted successfully"))

    except VerificationCode.DoesNotExist:
        return error_response(_("Invalid code"), 400)

    except Exception as e:
        return error_response(e.__str__())

    return error_response(_("Invalid code"), 400)



# Generated by Django 4.1.7 on 2023-08-16 05:18

from django.db import migrations, models
import shortuuid.main


class Migration(migrations.Migration):

    dependencies = [
        ('nets_core', '0003_userdevice'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='userdevice',
            name='device_id',
        ),

    ]



# Generated by Django 5.0.4 on 2024-05-07 18:03

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('nets_core', '0011_role_enabled'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='permission',
            name='created',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, verbose_name='Created'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='permission',
            name='updated',
            field=models.DateTimeField(auto_now=True, verbose_name='updated'),
        ),
        migrations.AddField(
            model_name='role',
            name='created',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, verbose_name='Created'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='role',
            name='updated',
            field=models.DateTimeField(auto_now=True, verbose_name='updated'),
        ),
        migrations.AddField(
            model_name='userrole',
            name='created',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, verbose_name='Created'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='userrole',
            name='updated',
            field=models.DateTimeField(auto_now=True, verbose_name='updated'),
        ),
        migrations.AlterField(
            model_name='userrole',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='roles', to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 4.1.7 on 2023-08-16 20:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('nets_core', '0008_verificationcode_ip_userfirebasenotification'),
    ]

    operations = [
        migrations.AddField(
            model_name='userdevice',
            name='ip',
            field=models.CharField(blank=True, max_length=250, null=True, verbose_name='IP'),
        ),
    ]



from functools import wraps
import logging

from django.apps import apps
from django.http.response import JsonResponse
from django.views.decorators.csrf import csrf_exempt
from django.utils.translation import gettext_lazy as _
from django.conf import settings

from nets_core.handlers import get_request_obj, request_params_handler
from nets_core.params import RequestParam
from nets_core.responses import permission_denied
from nets_core.utils import get_client_ip, check_perm



logger = logging.getLogger(__name__)

def request_handler(
    obj=None, 
    can_do: str=None, 
    perm_required: bool=False, 
    params: dict|list[RequestParam]={}, 
    optionals: dict={}, 
    allow_anonymous=False, 
    public=False,
    project_required=False,
    index_field: str = 'id'):
    """
        Decorator for request params handler
        check if customer is required, permissions and obj
        append required attributes to request object
        
        obj: model class
        can_do: list of permissions to check
        pem_required: if True, check if user has permission
        params: list of RequestParam objects
        optionals: dict of optional params
        allow_anonymous: if True, allow anonymous users
        public: if True, allow public users
        customer_required: if True, check if customer_id is present in request
            and retrieve customer from db append to request object
        index_field: field to use as index for obj
    """

    def decorator(view_func):
        @csrf_exempt
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):
            
            if request.user.is_anonymous and not public:
                return permission_denied()
            
            _params = params

            if isinstance(_params, list):
                # check if params are list of RequestParam
                # and convert to dict
                _params_dict = {}
                for p in _params:
                    _params_dict[p.key] = p
                _params = _params_dict
            
            request.project = None
            request.project_membership = None
            
            # if project_required:
            #     if not hasattr(settings, 'NETS_CORE_PROJECT_MODEL'):
            #         raise Exception('NETS_CORE_PROJECT_MODEL not set in settings')
            #     else:
            #         if not hasattr(request, 'project_id'):
            #             return JsonResponse({"res": 0, "message": _('project_id required')}, status=400)
                    
            #         project_model = apps.get_model(settings.NETS_CORE_PROJECT_MODEL)
            #         try:
            #             project = project_model.objects.get(id=request.project_id)
            #             request.project = project
            #         except project_model.DoesNotExist:
            #             return JsonResponse({"res": 0, "message": _('project not found')}, status=404)
                    
            #     if not hasattr(settings, 'NETS_CORE_PROJECT_MEMBER_MODEL'):
            #         raise Exception('NETS_CORE_PROJECT_MEMBER_MODEL not set in settings')
            #     else:
            #         project_member_model = apps.get_model(settings.NETS_CORE_PROJECT_MEMBER_MODEL)
            #         try:
            #             project_member = project_member_model.objects.get(user=request.user, project=project)
            #             request.project_membership = project_member
            #         except project_member_model.DoesNotExist:
            #             if perm_required:
            #                 return JsonResponse({"res": 0, "message": _('project member not found')}, status=404)
                        
            
                
            request.project_required = project_required
            request.public = public
            request.index_field = index_field
            request = request_params_handler(request, _params)
            if isinstance(request, JsonResponse):
                return request
            
            perm = public
            if can_do:
                # TODO: log permission check
                perm = check_perm(request.user, can_do, request.project)              
                
                # log_permission_check.delay(request.user.id, can_do, access=perm, customer_id=customer_id)

            if not perm and perm_required:
                return JsonResponse({"res": 0, "message": _('permission denied')}, status=403)
            
            request.perm = perm
            request.can_do = perm
            if obj:
                request = get_request_obj(request, obj)
                
                if isinstance(request, JsonResponse):
                    return request

            request.ip = get_client_ip(request)
            return view_func(request, *args, **kwargs)
        
        return _wrapped_view
    return decorator


# Generated by Django 4.1.7 on 2023-08-16 05:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('nets_core', '0005_userdevice_uuid'),
    ]

    operations = [
        migrations.AddField(
            model_name='verificationcode',
            name='device',
            field=models.CharField(blank=True, max_length=150, null=True, verbose_name='nets_core.UserDevice'),
        ),
    ]



from nets_core.firebase_messages import send_user_device_notification
from nets_core.models import NetsCoreBaseModel, VerificationCode
from django.db.models.signals import post_delete, post_save, pre_delete, post_migrate, post_init, pre_save
from django.dispatch import receiver
from nets_core.mail import send_email
from nets_core.models import EmailTemplate
from django.core.cache import cache
from django.utils.translation import gettext_lazy as _
from django.utils import timezone

import logging

logger = logging.getLogger(__name__)


@receiver(pre_save)
def pre_save_base_model_handler(sender, instance, **kwargs):
    if not instance.pk:
        return # created
    
    if issubclass(sender, NetsCoreBaseModel):
        # NetsCoreBaseModel implements updated_fields property to track fields that have changed
        # is a dict of field_name: [{'old': old_value, 'new': new_value, 'time': time}]
        untracked_fields = ['password', 'token', 'updated_fields']
        previous_instance = sender.objects.get(pk=instance.pk)
        for field in instance._meta.fields:
            if not hasattr(field, 'column') or field.column in ('created_at', 'updated_at'):
                continue
            # check if field is hasattr auto_now or auto_now_add or is BigAutoField
            if hasattr(field, 'auto_now') or hasattr(field, 'auto_now_add') or field.__class__.__name__ == 'BigAutoField':
                continue
            
            field_name = field.name
            if field_name in untracked_fields:
                continue
            
            if getattr(previous_instance, field_name) != getattr(instance, field_name):
                if not hasattr(instance, 'updated_fields'):
                    instance.updated_fields = {}
                if field_name not in instance.updated_fields:
                    instance.updated_fields[field_name] = []
                
                # ensure values are strings for JSON serialization
                try:
                    old_value = str(getattr(previous_instance, field_name))
                except:
                    old_value = None
                
                try:
                    new_value = str(getattr(instance, field_name))
                except:
                    new_value = None
                
                if not old_value and not new_value:
                    continue
                
                if old_value != new_value:
                    instance.updated_fields[field_name].append({
                        'old': old_value,
                        'new': new_value,
                        'time': str(timezone.now())
                    })


@receiver(post_save, sender=VerificationCode)
def send_verification_code_email(sender, instance, created, **kwargs):
    if created:
        # send email
        cache_token_key = instance.get_token_cache_key()
        button_link = {"url": "", "label": cache.get(cache_token_key)}
        template = None
        html = None
        email_template = (
            EmailTemplate.objects.filter(use_for="verification_code", enabled=True)
            .order_by("-created")
            .first()
        )

        if email_template:
            html = email_template.html_body
        else:
            template = "nets_core/email/verification_code.html"

        result = send_email(
            _("Verification code"),
            [instance.user.email],
            template,
            {"button_link": button_link, "user": instance.user},
            html=html,
            to_queued=False,
        )
        logger.info(
            f"Verification code email to {instance.user.email} with result {result}"
        )

    else:
        if instance.verified:
            # send notification to user about new login
            template = "nets_core/email/new_login.html"
            send_email(
                _("New login"),
                [instance.user.email],
                template,
                {"user": instance.user, "ip": str(instance.ip), "device": instance.device},
                to_queued=False,
            )

            message = _("New login to your account from ip address") + " " + instance.ip
            data = {"type": "login", "ip": str(instance.ip), "device": ""}
            if instance.device:
                message += " " + _("using device") + " " + instance.device.__str__()
                data["device"] = f"{instance.device.name}"
                data["device_id"] = f"{instance.device_id}"
            title = f"{_('New login')} {instance.user.username}"
            send_user_device_notification(
                instance.user, title=title, message=message, data=data
            )
